import abc
import numpy as np
from typing import Union, List, Tuple, Optional

from mentpy.states.mbqcstate import MBQCState


class BaseSimulator(abc.ABC):
    """Base class for simulators.

    Note
    ----
    This class should not be used directly. Instead, use one of the subclasses.

    Args
    ----
    mbqcstate: mp.MBQCState
        The MBQC state used for the simulation.
    input_state: np.ndarray
        The input state of the simulator.

    See Also
    --------
    :class:`mp.PatternSimulator`, :class:`mp.PennylaneSimulator`, :class:`mp.CirqSimulator`

    Group
    -----
    simulators
    """

    def __init__(self, mbqcstate: MBQCState, input_state: np.ndarray = None) -> None:
        self._mbqcstate = mbqcstate
        self._input_state = input_state

    @property
    def mbqcstate(self) -> MBQCState:
        """The MBQC state used for the simulation."""
        return self._mbqcstate

    @property
    def input_state(self) -> np.ndarray:
        """The input state of the simulator."""
        return self._input_state

    @input_state.setter
    def input_state(self, input_state: np.ndarray):
        """Sets the input state of the simulator."""
        self._input_state = input_state

    def __call__(
        self, angles: List[float], planes: Union[List[str], str] = "XY", **kwargs
    ):
        return self.measure_pattern(angles, planes, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} for {self.mbqcstate}"

    @abc.abstractmethod
    def measure(self, angle: float, plane: str = "XY", **kwargs):
        """Measures the state of the system.

        Args
        ----
        angle: float
            The angle of measurement.
        plane: str
            The plane of measurement.
        """
        pass

    @abc.abstractmethod
    def measure_pattern(
        self, angles: List[float], planes: Union[List[str], str] = "XY", **kwargs
    ) -> Tuple[List[int], np.ndarray]:
        """Measures the state of the system.

        Args
        ----
        angles: List[float]
            The angles of measurement.
        planes: List[str]
            The planes of measurement.
        """
        pass

    @abc.abstractmethod
    def reset(self, input_state=None):
        """Resets the simulator to the initial state."""
        pass



from typing import Union, List, Tuple, Optional
import numpy as np

from mentpy.states.mbqcstate import MBQCState
from mentpy.simulators.base_simulator import BaseSimulator


class CirqSimulator(BaseSimulator):
    """Simulator for measuring patterns of MBQC states.
    Note
    ----
    This is a placeholder for the Cirq simulator. It is not yet implemented.

    Args
    ----
    mbqcstate: mp.MBQCState
        The MBQC state used for the simulation.
    simulator: str
        The simulator to use.
    input_state: np.ndarray
        The input state of the simulator.

    See Also
    --------
    :class:`mp.PatternSimulator`, :class:`mp.PennylaneSimulator`

    Group
    -----
    simulators
    """

    def __init__(self, mbqcstate: MBQCState, input_state: np.ndarray = None) -> None:
        super().__init__(mbqcstate, input_state)

    def measure(self, angle: float, plane: str = "XY", **kwargs):
        raise NotImplementedError

    def measure_pattern(
        self, angles: List[float], planes: Union[List[str], str] = "XY", **kwargs
    ) -> Tuple[List[int], np.ndarray]:
        raise NotImplementedError

    def reset(self, input_state: np.ndarray = None):
        raise NotImplementedError



import numpy as np
from mentpy.operators import PauliOp
import networkx as nx
from typing import Optional, List

__all__ = ["GraphState", "entanglement_entropy"]


class GraphState(nx.Graph):
    """A graph state class that inherits from networkx.Graph.

    Examples
    --------
    Create a 1D cluster state :math:`|G>` of five qubits

    .. ipython:: python

        g = mp.GraphState()
        g.add_edges_from([(0,1), (1,2), (2,3), (3, 4)])
        print(g)

    See Also
    --------
    :class:`mentpy.state.mbqcstate.MBQCState`

    Group
    -----
    states
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"GraphState with {self.number_of_nodes()} nodes and {self.number_of_edges()} edges."

    def __len__(self):
        return self.number_of_nodes()

    def __eq__(self, other):
        return nx.is_isomorphic(self, other)
    
    def stabilizers(self):
        """
        Generate the stabilizers of a graph state.
        
        Examples
        --------
        Calculate the stabilizers of a 1D cluster state :math:`|G>` of five qubits

        .. ipython:: python
            :okwarning:

            g = mp.GraphState()
            g.add_edges_from([(0,1), (1,2), (2,3), (3, 4)])
            print(g.stabilizers())
        """
        return graphstate_stabs(self)


def lc_reduce(graph: GraphState):
    """Reduce graph state

    Group
    -----
    states
    """
    raise NotImplementedError


# TODO: Check if this is correct.
def entanglement_entropy(
    state: GraphState, subRegionA: List, subRegionB: Optional[List] = None
) -> float:
    """Calculate the entanglement entropy of a subregion of a graph state."""

    G = state.copy()

    # minimum_cut requires the capacity kwarg.
    nx.set_edge_attributes(G, 1, name="capacity")
    if subRegionB is None:
        subRegionB = set(state.nodes()) - set(subRegionA)

    # Allow subregions. These are merged into a supernode to calculate
    # the minimum cut between them.
    if isinstance(subRegionA, List):
        for v in subRegionA:
            G = nx.contracted_nodes(G, subRegionA[0], v)
        subRegionA = subRegionA[0]
    if isinstance(subRegionB, List):
        for v in subRegionA:
            G = nx.contracted_nodes(G, subRegionB[0], v)
        subRegionB = subRegionB[0]

    return nx.minimum_cut(G, subRegionA, subRegionB)[0]


def graphstate_stabs(graph: GraphState) -> List[PauliOp]:
    """Generate the stabilizers of a graph state.

    Examples
    --------
    Calculate the stabilizers of a 1D cluster state :math:`|G>` of five qubits
    
    Group
    -----
    states
    """
    z_mat = nx.adjacency_matrix(graph).todense()
    x_mat = np.eye(graph.number_of_nodes(), dtype=int)

    return PauliOp(np.hstack((x_mat, z_mat)))

    



from typing import Union, List, Tuple, Optional
import numpy as np

from mentpy.states.mbqcstate import MBQCState
from mentpy.simulators.base_simulator import BaseSimulator
from mentpy.simulators.pennylane_simulators import *
from mentpy.simulators.cirq_simulators import *

__all__ = ["PatternSimulator"]


class PatternSimulator:
    """Simulator for measuring patterns of MBQC states.

    Args
    ----
    mbqcstate: mp.MBQCState
        The MBQC state used for the simulation.
    simulator: str
        The simulator to use. Currently only 'pennylane-default.qubit' is supported.

    See Also
    --------
    :class:`mp.PennylaneSimulator`, :class:`mp.CirqSimulator`

    Group
    -----
    simulators
    """

    def __init__(
        self,
        mbqcstate: MBQCState,
        input_state: np.ndarray = None,
        simulator="pennylane-default.qubit",
        *args,
        **kwargs,
    ) -> None:

        supported_simulators = {
            "pennylane-default.qubit": PennylaneSimulator,
            "cirq": CirqSimulator,
        }

        if simulator not in supported_simulators:
            raise ValueError(
                f"Simulator {simulator} not supported. Supported simulators are {supported_simulators.keys()}"
            )

        if input_state is None:
            input_state = 1
            for i in range(len(mbqcstate.input_nodes)):
                input_state = np.kron(input_state, np.array([1, 1]) / np.sqrt(2))

        self.simulator = supported_simulators[simulator](
            mbqcstate, input_state, *args, **kwargs
        )

    def __getattr__(self, name):
        return getattr(self.simulator, name)

    def __call__(
        self, angles: List[float], planes: Union[List[str], str] = "XY", **kwargs
    ):
        return self.measure_pattern(angles, planes, **kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} ({self.simulator!r})"

    def measure(self, angle: float, plane: str = "XY", **kwargs):
        return self.simulator.measure(angle, plane, **kwargs)

    def measure_pattern(
        self, angles: List[float], planes: Union[List[str], str] = "XY", **kwargs
    ) -> Tuple[List[int], np.ndarray]:
        return self.simulator.measure_pattern(angles, planes, **kwargs)

    def reset(self, input_state: np.ndarray = None):
        return self.simulator.reset(input_state)



"""The graph_state module"""

from functools import cached_property
from typing import Optional, List, Tuple, Callable, Union, Any

import numpy as np
import scipy as scp
import networkx as nx

from mentpy.states.graphstate import GraphState
from mentpy.states.flow import find_flow, find_gflow

__all__ = ["MBQCState", "draw"]


class MBQCState:
    r"""The MBQCGraph class that deals with operations and manipulations of graph states
    Args
    ----
    graph: mp.GraphState
        The graph state of the MBQC state.
    input_nodes: list
        The input nodes of the MBQC state.
    output_nodes: list
        The output nodes of the MBQC state.
    wires: Optional[list]
        The wires of the MBQC state.

    Examples
    --------
    Create a 1D cluster state :math:`|G>` of five qubits

    .. ipython:: python

        g = mp.GraphState()
        g.add_edges_from([(0,1), (1,2), (2,3), (3, 4)])
        state = mp.MBQCState(g, input_nodes=[0], output_nodes=[4])


    See Also
    --------
    :class:`mp.GraphState`

    Group
    -----
    states
    """

    def __init__(
        self,
        graph: GraphState,
        input_nodes: np.ndarray = np.array([]),
        output_nodes: np.ndarray = np.array([]),
        flow: Optional[Callable] = None,
        partial_order: Optional[callable] = None,
        measurement_order: Optional[List[int]] = None,
        gflow: Optional[Callable] = None,
    ) -> None:
        """Initializes a graph state"""
        self._graph = graph

        # check input and output nodes are in graph. If not, raise error with the node(s) that are not in the graph
        if not all([v in self.graph.nodes for v in input_nodes]):
            raise ValueError(
                f"Input nodes {input_nodes} are not in the graph. Graph nodes are {self.graph.nodes}"
            )
        if not all([v in self.graph.nodes for v in output_nodes]):
            raise ValueError(
                f"Output nodes {output_nodes} are not in the graph. Graph nodes are {self.graph.nodes}"
            )
        self._input_nodes = input_nodes
        self._output_nodes = output_nodes

        if (flow is None) or (partial_order is None):
            flow, partial_order = find_flow(graph, input_nodes, output_nodes)

        self._flow = flow
        self._partial_order = partial_order

        if gflow is None:
            if flow is None:
                gflow, gpartial_order, g_func = find_gflow(
                    graph, input_nodes, output_nodes
                )
            else:
                gflow, gpartial_order, g_func = flow, partial_order, flow

        self.gflow = gflow

        if measurement_order is None and flow is not None:
            measurement_order = self.calculate_order()

        self._measurement_order = measurement_order

    def __repr__(self) -> str:
        """Return the representation of the current graph state"""
        return (
            f"MBQCState with {self.graph.number_of_nodes()} nodes and "
            f"{self.graph.number_of_edges()} edges"
        )

    def __len__(self) -> int:
        """Return the number of nodes in the GraphStateCircuit"""
        return len(self.graph)
    
    # if an attribute is not found, look for it in the graph
    def __getattr__(self, name):
        return getattr(self.graph, name)

    @property
    def graph(self) -> GraphState:
        r"""Return the graph of the resource state."""
        return self._graph

    @property
    def input_nodes(self) -> np.ndarray:
        r"""Return the input nodes of the MBQC circuit."""
        return self._input_nodes

    @property
    def output_nodes(self) -> np.ndarray:
        r"""Return the output nodes of the MBQC circuit."""
        return self._output_nodes

    @property
    def flow(self) -> Callable:
        r"""Return the flow function of the MBQC circuit."""
        return self._flow

    @property
    def partial_order(self) -> Callable:
        r"""Return the partial order function of the MBQC circuit."""
        return self._partial_order

    @property
    def measurement_order(self) -> List[int]:
        r"""Return the measurement order of the MBQC circuit."""
        return self._measurement_order

    @measurement_order.setter
    def measurement_order(self, measurement_order: List[int]) -> None:
        r"""Set the measurement order of the MBQC circuit."""
        if not _check_measurement_order(measurement_order, self.partial_order):
            raise ValueError(f"Invalid measurement order {measurement_order}.")
        self._measurement_order = measurement_order

    @cached_property
    def outputc(self):
        r"""Returns :math:`O^c`, the complement of output nodes."""
        return [v for v in self.graph.nodes() if v not in self.output_nodes]

    @cached_property
    def inputc(self):
        r"""Returns :math:`I^c`, the complement of input nodes."""
        return [v for v in self.graph.nodes() if v not in self.input_nodes]

    def calculate_order(self):
        r"""Returns the order of the measurements"""
        n = len(self.graph)
        mat = np.zeros((n, n))

        for indi, i in enumerate(list(self.graph.nodes())):
            for indj, j in enumerate(list(self.graph.nodes())):
                if self.partial_order(i, j):
                    mat[indi, indj] = 1

        sum_mat = np.sum(mat, axis=1)
        order = np.argsort(sum_mat)[::-1]

        # turn order into labels of graph
        order = [list(self.graph.nodes())[i] for i in order]

        return order

    def add_edge(self, u, v):
        r"""Adds an edge between nodes u and v"""
        self.graph.add_edge(u, v)
        # try resetting self with new graph, if it fails, remove the edge
        try:
            self.__init__(self.graph, self.input_nodes, self.output_nodes)
        except Exception as e:
            self.graph.remove_edge(u, v)
            raise ValueError(f"Cannot add edge between {u} and {v}.\n" + str(e))


def _check_measurement_order(measurement_order: List, partial_order: Callable):
    r"""Checks that the measurement order is valid"""
    for i in range(len(measurement_order)):
        for j in range(i + 1, len(measurement_order)):
            if partial_order(measurement_order[i], measurement_order[j]):
                return False
    return True


def merge(
    state1: MBQCState, state2: MBQCState, indices_tuple: List[Tuple]
) -> MBQCState:
    """Merge two graph states into a larger graph state

    Group
    -----
    states
    """
    raise NotImplementedError


def draw(state: Union[MBQCState, GraphState], fix_wires=None, **kwargs):
    """Draws mbqc circuit with flow.

    Group
    -----
    states
    """
    node_colors = {
        i: "#FFBD59"
        if i in set(state.inputc).intersection(set(state.outputc))
        else "#CCCCCC"
        for i in state.graph.nodes()
    }
    # options = {'node_color': '#FFBD59'}
    options = {"node_color": [node_colors[node] for node in state.graph.nodes()]}
    options.update(kwargs)

    if isinstance(state, GraphState):
        nx.draw(state, **options)

    elif isinstance(state, MBQCState):
        fixed_nodes = state.input_nodes + state.output_nodes
        position_xy = {}
        for indx, p in enumerate(state.input_nodes):
            position_xy[p] = (0, -2 * indx)

        separation = len(state.outputc) // len(state.output_nodes)
        if fix_wires is not None:
            for wire in fix_wires:
                if len(wire) + 2 > separation:
                    separation = len(wire) + 2
        for indx, p in enumerate(state.output_nodes):
            position_xy[p] = (2 * (separation), -2 * indx)

        if fix_wires is not None:
            x = [list(x) for x in fix_wires]
            fixed_nodes += sum(x, [])
            for indw, wire in enumerate(fix_wires):
                for indx, p in enumerate(wire):
                    position_xy[p] = (2 * (indx + 1), -2 * indw)

        node_pos = nx.spring_layout(
            state.graph, pos=position_xy, fixed=fixed_nodes, k=1 / len(state.graph)
        )
        nx.draw(state.graph, pos=node_pos, **options)
        if state.flow is not None:
            nx.draw(_graph_with_flow(state), pos=node_pos, **options)


def _graph_with_flow(state):
    """Return digraph with flow (but does not have all CZ edges!)"""
    g = state.graph
    gflow = nx.DiGraph()
    gflow.add_nodes_from(g.nodes)
    for node in state.outputc:
        gflow.add_edge(node, state.flow(node))
    return gflow



from typing import Union, List, Tuple, Optional

import numpy as np

from mentpy.simulators.base_simulator import BaseSimulator
from mentpy.states.mbqcstate import MBQCState

import pennylane as qml
import networkx as nx

__all__ = ["PennylaneSimulator"]


class PennylaneSimulator(BaseSimulator):
    """Simulator for measuring patterns of MBQC states.

    Args
    ----
    mbqcstate: mp.MBQCState
        The MBQC state used for the simulation.
    input_state: np.ndarray
        The input state of the simulator.

    See Also
    --------
    :class:`mp.PatternSimulator`, :class:`mp.CirqSimulator`

    Group
    -----
    simulators
    """

    def __init__(
        self, mbqcstate: MBQCState, input_state: np.ndarray, *args, **kwargs
    ) -> None:
        self.circuit = graphstate_to_circuit(mbqcstate)
        super().__init__(mbqcstate, input_state)

    def measure(self, angle: float, plane: str = "XY"):
        raise NotImplementedError

    def measure_pattern(
        self, angles: List[float], planes: Union[List[str], str] = "XY", **kwargs
    ) -> Tuple[List[int], np.ndarray]:

        # TODO: Implement this
        if planes != "XY":
            raise NotImplementedError

        return self.circuit(angles, st=self.input_state, **kwargs)

    def reset(self, input_state=None):
        if input_state is None:
            input_state = self.input_state
        self.input_state = input_state


def graphstate_to_circuit(gsc):
    """Converts a MBQCState to a PennyLane circuit."""
    gr = gsc.graph
    N = gr.number_of_nodes()
    dev = qml.device("default.qubit", wires=N)

    @qml.qnode(dev)
    def circuit(param, output="density", st=None):
        if output != "density":
            assert (
                len(param) == N
            ), f"Length of param is {len(param)}, but expected {N}."
        else:
            assert (
                len(param) == N - 2
            ), f"Length of param is {len(param)}, but expected {N-2}."
        input_v = st
        qml.QubitStateVector(input_v, wires=gsc.input_nodes)
        for j in gsc.inputc:
            qml.Hadamard(j)
        for i, j in gr.edges():
            qml.CZ(wires=[i, j])

        topord_no_output = [
            x for x in gsc.measurement_order if (x not in gsc.output_nodes)
        ]
        for indx, p in zip(topord_no_output, param[: len(gsc.outputc)]):
            qml.RZ(p, wires=indx)
            qml.Hadamard(wires=indx)
            m_0 = qml.measure(indx)
            qml.cond(m_0, qml.PauliX)(wires=gsc.flow(indx))
            for neigh in gr.neighbors(gsc.flow(indx)):
                if neigh != indx and (
                    gsc.measurement_order.index(neigh)
                    > gsc.measurement_order.index(indx)
                ):
                    qml.cond(m_0, qml.PauliZ)(wires=neigh)

        if output == "expval":
            for indx, p in zip(gsc.output_nodes, param[len(gsc.outputc) :]):
                qml.RZ(p, wires=indx)
            return [qml.expval(qml.PauliX(j)) for j in gsc.output_nodes]
        elif output == "sample":
            for indx, p in zip(gsc.output_nodes, param[len(gsc.outputc) :]):
                qml.RZ(p, wires=indx)
            return [qml.sample(qml.PauliX(j)) for j in gsc.output_nodes]
        elif output == "density":
            return qml.density_matrix(gsc.output_nodes)

    return circuit



"""
This is the common_ansatz module. 
It has several common ansatzes that can be used for MBQC algorithms
"""

from typing import List
from mentpy.states.graphstate import GraphState
from mentpy.states.mbqcstate import MBQCState

__all__ = ["linear_cluster", "many_wires"]


def linear_cluster(n) -> MBQCState:
    r"""Returns a linear cluster state of n qubits.
    Args
    ----
    n: int
        The number of qubits in the cluster state.

    Returns
    -------
    The linear cluster state of n qubits.

    Examples
    --------
    Create a 1D cluster state :math:`|G>` of five qubits
    .. ipython:: python

        g = mp.templates.linear_cluster(5)
        print(g)

    Group
    -----
    states
    """
    g = GraphState()
    g.add_edges_from([(i, i + 1) for i in range(n - 1)])
    gs = MBQCState(g, input_nodes=[0], output_nodes=[n - 1])
    return gs


def many_wires(n_wires: List) -> MBQCState:
    r"""Returns a graph state with many wires.
    Args
    ----
    n_wires: List
        A list of the number of qubits in each wire.

    Returns
    -------
    The graph state with many wires.

    Examples
    --------
    Create a graph state with three wires of 2, 3, and 4 qubits respectively
    .. ipython:: python

        g = mp.templates.many_wires([2, 3, 4])
        print(g)

    Group
    -----
    states
    """
    # require n_wires to be a list of integers greater than 1
    if not all([isinstance(n, int) and n > 1 for n in n_wires]):
        raise ValueError("n_wires must be a list of integers greater than 1")

    g = GraphState()
    for i, n in enumerate(n_wires):
        g.add_edges_from(
            [(j + sum(n_wires[:i]), j + sum(n_wires[:i]) + 1) for j in range(n - 1)]
        )

    # input nodes are the first qubit in each wire and output nodes are the last qubit in each wire
    gs = MBQCState(
        g,
        input_nodes=[sum(n_wires[:i]) for i in range(len(n_wires))],
        output_nodes=[sum(n_wires[: i + 1]) - 1 for i in range(len(n_wires))],
    )
    return gs


def grid_cluster(n, m) -> MBQCState:
    r"""Returns a grid cluster state of n x m qubits.
    Args
    ----
    n: int
        The number of rows in the cluster state.
    m: int
        The number of columns in the cluster state.

    Returns
    -------
    The grid cluster state of n x m qubits.

    Examples
    --------
    Create a 2D cluster state :math:`|G>` of five qubits

    .. ipython:: python

        g = mp.templates.grid_cluster(2, 3)
        print(g)

    Group
    -----
    states
    """
    g = GraphState()
    # add edges between rows
    n_wires = [n] * m
    for i, n in enumerate(n_wires):
        g.add_edges_from(
            [(j + sum(n_wires[:i]), j + sum(n_wires[:i]) + 1) for j in range(n - 1)]
        )

    # add edges between columns
    for i in range(m - 1):
        g.add_edges_from([(i * n + j, (i + 1) * n + j) for j in range(n)])

    gs = MBQCState(
        g,
        input_nodes=[sum(n_wires[:i]) for i in range(len(n_wires))],
        output_nodes=[sum(n_wires[: i + 1]) - 1 for i in range(len(n_wires))],
    )
    return gs



import numpy as np
import cirq

CNOT = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])

SWAP = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])

CS = np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1j]])

H = np.array([[1, 1], [1, -1]]) / np.sqrt(2)

PauliX = np.array([[0, 1], [1, 0]])

PauliY = np.array([[0, -1j], [1j, 0]])

PauliZ = np.array([[1, 0], [0, -1]])

T = np.array([[1, 0], [0, np.exp(1j * np.pi / 4)]])


def random_su(n_qubits: int):
    """Returns a random special unitary in ``n_qubits`` sampled from the Haar distribution."""
    return cirq.testing.random_special_unitary(dim=int(2**n_qubits))



"""Helper functions for linear algebra used for testing LC equivalence.

TAKEN FROM: https://github.com/XanaduAI/flamingpy/blob/main/flamingpy/utils/linalg.py
"""

import numpy as np
import networkx as nx


def reduce_RREform_mod2(M, max_cols=None):
    """Put a binary matrix into row reduced echelon form modulo 2, up to a
    maximum number of columns given by max_cols.
    Args:
        M (numpy.array): the matrix to reduce.
        max_cols (int): the maximum number of columns of the input array to reduce.
    Returns:
        (numpy.array, int): the reduced matrix and the the index of pivot column.
    """
    # number of columns to apply row reduction
    max_cols = M.shape[1] if max_cols is None else max_cols
    # row reduced matrix
    R = M.copy()
    # pivot index
    p = 0
    # perform row reduction
    for j in range(max_cols):
        # look for pivot column `j` at or below row `p`
        index = np.nonzero(R[p:, j])[0]
        if index.size == 0:
            continue
        # pivot row
        i = p + index[0]
        # interchange `p` and `i`
        R[[p, i], :] = R[[i, p], :]
        # set pivot to 1
        R[p, :] = R[p, :] / R[p, j]
        # add zeros above and below the pivot
        index = np.nonzero(R[:, j])[0].tolist()
        index.remove(p)
        R[index, :] = (R[index, :] - np.multiply.outer(R[index, j], R[p, :])) % 2
        # increment pivot index until final row is reached
        p += 1
        if p == R.shape[0]:
            break
    return R, p


def lc_constraint_system(G, H):
    """Build the constraint for LC-equivalence of two adjacency matrices.
    Construct a binary system of equations that two adjacency matrices
    G and H must satisfy for equivalence through local complementations.
    Args:
        G, H (numpy.array): n x n adjacency matrices.
    Returns:
        numpy.array: an array of shape n^2 x 4n representing a system of n^2
            binary linear equations with 4n unknowns.
    """
    # set the number of qubits
    n = np.shape(G)[0]
    # define empty block matrix
    M = []
    for j in range(n):
        # A block
        A = np.diag(G[j])
        # B block
        B = np.zeros((n, n))
        B[j, j] = 1
        # C block
        C = np.array(
            [[G[i, j] * H[i, k] for i in range(n)] for k in range(n)], dtype=int
        )
        # D block
        D = np.zeros((n, n), dtype=int)
        for k in range(n):
            D[k, j] = H[j, k]
        # add row block to M
        M.append([A, B, C, D])
    # define numpy block matrix for system of constraints
    return np.block(M).astype(int)


def nullspace_basis(M):
    """Return the nullspace basis of matrix M.
    Construct an array whose rows are binary basis vectors of the right
    nullspace of input matrix array.
    Args:
        M (numpy.array): a binary matrix.
    Returns:
        numpy.array: an array whose rows are basis vectors of the right
            nullspace of M.
    """
    # find left Null space of transposed system, which is equal to right null space
    M_transposed = M.T
    m_rows, n_cols = M_transposed.shape
    # construct augmented block matrix A = [M | I]
    A = np.concatenate((M_transposed, np.eye(m_rows, dtype=int)), axis=-1)
    # row reduce left M block of augmented matrix
    R, p = reduce_RREform_mod2(A, max_cols=n_cols)
    N = R[p:, n_cols:]
    basis, _ = reduce_RREform_mod2(N)
    return basis


def search_nullspace(basis):
    """Check if the nullspace satisfies the determinant constraints.
     Search through sums of pairs of basis vectors of the nullspace and
     see if any satisfy the determinant constraints. If no solution is found,
     output None. If a solution is found, return a vector specifying the local
     Clifford in the form:
        (a_1, a_2, ..., a_n, b_1, b_2, ..., b_n, c_1, c_2, ..., c_n, d_1, d_2, ..., d_n),
     where n is the number of nodes of the graph.
    Args:
        basis (numpy.array): an array whose rows are basis vectors of the nullspace
    Returns:
        NoneType or numpy.array: None if no solution, or the solution array.
    """
    # number of basis vectors of null space
    d = basis.shape[0]
    # number of nodes/qubits
    n = int(basis.shape[1] / 4)
    # boolean for LC equivalence
    equivalent = False
    # array to store potential solutions
    sols = []
    # search through distinct pairs of basis vectors
    for i in range(d):
        for j in range(i):
            # boolean for checking solution constraints
            sat = True
            # potential solution as sum (modulo 2) of basis pair
            sol = np.vectorize(lambda x: x % 2)(np.add(basis[i], basis[j]))
            # store potential solution
            sols.append(sol)
            for k in range(n):
                # check the determinants (modulo 2)
                det = (sol[k] * sol[k + 3 * n] + sol[k + n] * sol[k + 2 * n]) % 2
                if det != 1:
                    # constraints not satisfied
                    sat = False
                    # remove the false solution
                    sols.pop()
                    break
            # if solution found, set equivalent to True
            if sat:
                # equivalence is satisfied
                equivalent = sat
                break
        # end search loop
        if equivalent:
            break
    # if no solution found, return None
    if not equivalent:
        return None
    # if solution found, return clifford in vector form
    solution_vector = np.array(sols[0]).astype(int)
    return solution_vector


def clifford_vec_to_tensors(vec):
    """Convert a local Clifford gate on n qubits to a list of n single-qubit
    Cliffords.
    Take a local Clifford operation on n qubits in the form
    (a_1, a_2, ..., a_n, b_1, b_2, ..., b_n, c_1, c_2, ..., c_n, d_1, d_2, ..., d_n),
    and return a list of single-qubit Clifford gates as a list of 2x2 arrays, where the array
    at index k corresponds to the Clifford acting on qubit k given as:
        [a_k, b_k]
        [c_k, d_k].
    Args:
        vec (numpy.array): a vector of size 4n specifying a local Clifford gate on n qubits.
    Returns:
        list[numpy.array]: n 2x2 arrays representing single-qubit local Clifford gates.
    """
    # number of nodes/qubits
    n = int(len(vec) / 4)
    # initialize empty list
    local_clifford_list = []
    for i in range(n):
        single_qubit_clifford = np.array(
            [[vec[i], vec[i + n]], [vec[i + 2 * n], vec[i + 3 * n]]]
        )
        local_clifford_list.append(single_qubit_clifford)
    return local_clifford_list


def clifford_vec_to_global(vec):
    """Convert a local Clifford gate on n qubits to a local Clifford gate on
    all n qubits.
    Take a vector corresponding to a local Clifford gate on n qubits,
    (a_1, a_2, ..., a_n, b_1, b_2, ..., b_n, c_1, c_2, ..., c_n, d_1, d_2, ..., d_n),
    and return a single 2n x 2n array representing a local Clifford acting on
    all n qubits, given as the block matrix:
            [A, B]
            [C, D]
        where each block is a n x n diagonal matrix:
            A = diag(a_1, a_2, ..., a_n)
            B = diag(b_1, b_2, ..., b_n)
            C = diag(c_1, c_2, ..., c_n)
            D = diag(d_1, d_2, ..., d_n).
    Args:
        vec (numpy.array): a vector of size 4n specifying a local Clifford on n qubits.
    Returns:
        numpy.array:  a 2n x 2n numpy array representing a local
            Clifford acting on all n qubits.
    """
    # number of nodes/qubits
    n = int(len(vec) / 4)
    blocks = [np.diag([vec[i + k * n] for i in range(n)]) for k in range(4)]
    return np.block([[blocks[0], blocks[1]], [blocks[2], blocks[3]]])


def are_lc_equivalent(graph1, graph2, clifford_form="tensor"):
    """Check if two EGraphs are LC equivalent, and return the Clifford
    operation if so. Implemented as in arXiv:quant-ph/0405023.
    Args:
        graph1: the initial graph to check Clifford equivalence with.
        graph2: the graph to check Clifford equivalence against.
        clifford_form: a string describing the output form of local Clifford operation, if
            it exists.
            If 'tensor' (default), produce a list of length n of 2x2 numpy arrays corresponding to
                single-qubit tensor factors.
            If 'global', return a single 2nx2n numpy array corresponding to the global operator
                acting on all n qubits.
    Returns:
        (bool, numpy.array): whether the states are LC equivalent, and if they are, the local
            Clifford output according to 'clifford_form' specification.
    """
    # get adjacency matrices of input graphs
    G = nx.adjacency_matrix(graph1.graph)
    H = nx.adjacency_matrix(graph2.graph)

    # check adjacency matrices for same non-empty square shape
    if np.shape(G) == (0, 0) or np.shape(H) == (0, 0):
        # raise ValueError('Input Graphs must be non-empty')
        # Mark this case as not equivalent
        return False, None
    if np.shape(G) != np.shape(H):
        # raise ValueError('Input Graphs must have same number of nodes.')
        # Mark this case as not equivalent
        return False, None
    # adjacency matrices must be square
    if np.shape(G)[0] != np.shape(G)[1]:
        raise ValueError("Input matrices must be square.")

    # perform algorithm to search for solution
    #
    # construct system of constraints for two adjacency matrices G and H
    system = lc_constraint_system(G, H)
    # construct nullspace basis of system of constraints
    nullspace = nullspace_basis(system)
    # search nullspace for solution vector
    solution_vector = search_nullspace(nullspace)

    # if graphs are not equivalent set clifford_output to None
    if solution_vector is None:
        equivalent = False
        clifford_output = None
    # if graphs are equivalent and clifford_form == "tensor", convert clifford_output to
    # tensor factors
    elif clifford_form == "tensor":
        equivalent = True
        clifford_output = clifford_vec_to_tensors(solution_vector)
    # if graphs are equivalent and clifford_form == "global", convert clifford_output to
    # global form
    elif clifford_form == "global":
        equivalent = True
        clifford_output = clifford_vec_to_global(solution_vector)
    # return equivalence and clifford_output
    return equivalent, clifford_output



"""
This module provides the functionalities to define graph states
"""

from .graphstate import *
from .mbqcstate import *
from .templates import *

__all__ = ["GraphState", "MBQCState", "draw", "templates"]



import numpy as np


def estimate_gradient(f, x, h=1e-5, type="central"):

    if type not in ["central", "forward", "backward"]:
        raise UserWarning(
            f"Expected type to be 'central', 'forward', or 'backward' but {type} was given"
        )

    grad = np.zeros(len(x))
    for i in range(len(x)):
        if type == "central":
            grad[i] = (f(x + h * np.eye(len(x))[i]) - f(x - h * np.eye(len(x))[i])) / (
                2 * h
            )
        elif type == "forward":
            grad[i] = (f(x + h * np.eye(len(x))[i]) - f(x)) / h
        elif type == "backward":
            grad[i] = (f(x) - f(x - h * np.eye(len(x))[i])) / h
    return grad


def estimate_hessian(f, x, h=1e-5, type="central"):

    if type not in ["central", "forward", "backward"]:
        raise UserWarning(
            f"Expected type to be 'central', 'forward', or 'backward' but {type} was given"
        )

    hess = np.zeros((len(x), len(x)))
    for i in range(len(x)):
        for j in range(len(x)):
            if type == "central":
                hess[i, j] = (
                    f(x + h * np.eye(len(x))[i] + h * np.eye(len(x))[j])
                    - f(x + h * np.eye(len(x))[i] - h * np.eye(len(x))[j])
                    - f(x - h * np.eye(len(x))[i] + h * np.eye(len(x))[j])
                    + f(x - h * np.eye(len(x))[i] - h * np.eye(len(x))[j])
                ) / (4 * h**2)
            elif type == "forward":
                hess[i, j] = (
                    f(x + h * np.eye(len(x))[i] + h * np.eye(len(x))[j])
                    - f(x + h * np.eye(len(x))[i])
                    - f(x + h * np.eye(len(x))[j])
                    + f(x)
                ) / h**2
            elif type == "backward":
                hess[i, j] = (
                    f(x)
                    - f(x - h * np.eye(len(x))[i])
                    - f(x - h * np.eye(len(x))[j])
                    + f(x - h * np.eye(len(x))[i] - h * np.eye(len(x))[j])
                ) / h**2
    return hess



import numpy as np
from mentpy.gradients import estimate_gradient

from mentpy.optimizers.base_optimizer import BaseOptimizer


class AdamOptimizer(BaseOptimizer):
    """Class for the Adam optimizer.

    Parameters
    ----------
    step_size : float, optional
        The step size of the optimizer, by default 0.1
    b1 : float, optional
        The first moment decay rate, by default 0.9
    b2 : float, optional
        The second moment decay rate, by default 0.999
    eps : float, optional
        A small number to avoid division by zero, by default 10**-8

    Examples
    --------
    Create an Adam optimizer

    .. ipython:: python

        opt = mp.optimizers.AdamOptimizer()
        print(opt)


    See Also
    --------
    :class:`mp.optimizers.SGDOptimizer`
    
    Group
    -----
    optimizers
    """

    def __init__(self, step_size=0.1, b1=0.9, b2=0.999, eps=10**-8) -> None:
        """Initialize the Adam optimizer."""
        self.step_size = step_size
        self.b1 = b1
        self.b2 = b2
        self.eps = eps

    def optimize(self, f, x0, num_iters=100, callback=None, verbose=False, **kwargs):
        """Optimize a function f using the Adam optimizer."""
        m = np.zeros(len(x0))
        v = np.zeros(len(x0))
        x = x0

        for i in range(num_iters):
            # Adam Optimizer
            g = estimate_gradient(f, x, **kwargs)
            m = (1 - self.b1) * g + self.b1 * m  # First  moment estimate.
            v = (1 - self.b2) * (g**2) + self.b2 * v  # Second moment estimate.
            mhat = m / (1 - self.b1 ** (i + 1))  # Bias correction.
            vhat = v / (1 - self.b2 ** (i + 1))
            x = x - self.step_size * mhat / (np.sqrt(vhat) + self.eps)
            if callback is not None:
                callback(x, i)
            if verbose:
                print(f"Iteration {i+1} of {num_iters}: {x} with value {f(x)}")
        return x

    def update_step_size(self, x, i, factor=0.99):
        """Update the step size of the optimizer."""
        self.step_size = self.step_size * factor

    def optimize_and_gradient_norm(
        self, f, x0, num_iters=100, callback=None, verbose=False, **kwargs
    ):
        """Optimize a function f using the Adam optimizer."""
        m = np.zeros(len(x0))
        v = np.zeros(len(x0))
        x = x0
        norm = np.zeros(num_iters)

        for i in range(num_iters):
            # Adam Optimizer
            g = estimate_gradient(f, x, **kwargs)
            m = (1 - self.b1) * g + self.b1 * m
            v = (1 - self.b2) * (g**2) + self.b2 * v
            mhat = m / (1 - self.b1 ** (i + 1))
            vhat = v / (1 - self.b2 ** (i + 1))
            x = x - self.step_size * mhat / (np.sqrt(vhat) + self.eps)
            norm[i] = np.linalg.norm(g)
            if callback is not None:
                callback(x, i)
            if verbose:
                print(f"Iteration {i+1} of {num_iters}: {x} with value {f(x)}")

        return x, norm

    def reset(self):
        """Reset the optimizer."""
        pass



from mentpy.optimizers.base_optimizer import BaseOptimizer
from mentpy.gradients import estimate_gradient

import numpy as np


class SGDOptimizer(BaseOptimizer):
    """Class for the SGD optimizer.

    Args
    ----
    step_size : float, optional
        The step size of the optimizer, by default 0.1
    momentum : float, optional
        The momentum of the optimizer, by default 0.9
    nesterov : bool, optional
        Whether to use Nesterov momentum, by default False

    Examples
    --------
    Create an SGD optimizer

    .. ipython:: python

        opt = mp.optimizers.SGDOptimizer()
        print(opt)

    See Also
    --------
    :class:`mp.optimizers.AdamOptimizer`

    Group
    -----
    optimizers
    """

    def __init__(self, step_size=0.1, momentum=0.9, nesterov=False) -> None:
        """Initialize the SGD optimizer."""
        self.step_size = step_size
        self.momentum = momentum
        self.nesterov = nesterov

    def optimize(self, f, x0, num_iters=100, callback=None, verbose=False, **kwargs):
        """Optimize a function f using the SGD optimizer."""
        v = np.zeros(len(x0))
        x = x0

        for i in range(num_iters):
            # SGD Optimizer
            g = estimate_gradient(f, x, **kwargs)
            v = self.momentum * v - self.step_size * g
            if self.nesterov:
                x = x + self.momentum * v - self.step_size * g
            else:
                x = x + v
            if callback is not None:
                callback(x, i)
            if verbose:
                print(f"Iteration {i+1} of {num_iters}: {x} with value {f(x)}")
        return x

    def update_step_size(self, x, i, factor=0.99):
        """Update the step size of the optimizer."""
        self.step_size = self.step_size * factor

    def optimize_and_gradient_norm(
        self, f, x0, num_iters=100, callback=None, verbose=False, **kwargs
    ):
        """Optimize a function f using the SGD optimizer."""
        v = np.zeros(len(x0))
        x = x0
        norm = []
        for i in range(num_iters):
            # SGD Optimizer
            g = estimate_gradient(f, x, **kwargs)
            norm.append(np.linalg.norm(g))
            v = self.momentum * v - self.step_size * g
            if self.nesterov:
                x = x + self.momentum * v - self.step_size * g
            else:
                x = x + v
            if callback is not None:
                callback(x, i)
            if verbose:
                print(f"Iteration {i+1} of {num_iters}: {x} with value {f(x)}")
        return x, norm

    def reset(self, *args, **kwargs):
        pass



"""
The Measurement-Based Quantum computing simulator.
"""

from .states import *
from .operators import *
from .simulators import *


# from . import measurement

from . import gradients
from . import optimizers
from . import utils

__version__ = "0.0.0"
__version_info__ = (0, 0, 0)



import abc


class BaseOptimizer(abc.ABC):
    """Base class for optimizers.

    Note
    ----
    This class should not be used directly. Instead, use one of the subclasses.

    See Also
    --------
    :class:`mp.optimizers.SGDOptimizer`, :class:`mp.optimizers.AdamOptimizer`

    Group
    -----
    optimizers
    """

    def __init__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def optimize(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def optimize_and_gradient_norm(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def reset(self, *args, **kwargs):
        pass



from mentpy.states import MBQCState

import networkx as nx
import itertools


class FlowSpace:
    r"""The flow space graph of a MBQCGraph.

    Each node corresponds to a possible graph over ``len(graph_state)`` qubits.
    Each edge between nodes represent going from one graph to another via adding or removing edges.

    .. note::  ``flow_space()`` will only work for MBQCGraph with less
    than 8 qubits.
    """

    def __init__(self, graph_state: MBQCState, allow_any_size_graphs: bool = False):
        """Creates the flow graph space of a graph state circuit."""

        if len(graph_state) > 7 and (not allow_any_size_graphs):
            raise UserWarning(
                f"Expected a graph_state of size 7 or less, but {len(graph_state)} "
                "was given."
            )

        self.number_nodes = len(graph_state)
        self.input_nodes = graph_state.input_nodes
        self.output_nodes = graph_state.output_nodes
        total_graph, wf_graph, wof_graph = self.all_graphs_graph()
        self.total_graph_space = total_graph
        self.flow_graph_space = wf_graph
        self.no_flow_graph_space = wof_graph

    def __repr__(self) -> str:
        r"""Returns the representation of the flow space"""
        return (
            f"Flow space over {self.number_nodes} nodes with {self.input_nodes} input nodes "
            f"and {self.output_nodes} output nodes."
        )

    def generator_all_graphs(self):
        """Returns a generator that generates all possible graphs for :math:`n` ordered nodes."""
        n = self.number_nodes
        total_n = int(n * (n - 1) / 2)
        completegraph = nx.complete_graph(n)
        its = itertools.product([0, 1], repeat=total_n)
        for j in its:
            g = nx.Graph()
            g.add_nodes_from(list(range(n)))
            edgs = [edg for i, edg in enumerate(list(completegraph.edges())) if j[i]]
            g.add_edges_from(edgs)
            yield g

    def all_graphs_graph(self):
        """Returns a tuple with three graphs.

        The first graph is the graph corresponding to all possible graphs.
        The second is the subgraph of all graphs with flow. The third is the
        subgraph of all graphs without flow."""

        graphs_list = list(self.generator_all_graphs())
        graph_space = nx.Graph()
        wflow = []
        woflow = []
        for ind, g in enumerate(graphs_list):
            gs = MBQCState(
                g, input_nodes=self.input_nodes, output_nodes=self.output_nodes
            )

            fl = gs.flow

            if fl is not None:
                graph_space.add_node(ind, flow=True, mbqc_circuit=gs)
                wflow.append(ind)
            else:
                graph_space.add_node(ind, flow=False, mbqc_circuit=gs)
                woflow.append(ind)

        for ind1, g1 in enumerate(graphs_list):
            for ind2, g2 in enumerate(graphs_list):
                if ind1 < ind2:
                    if len(set(g2.edges()).symmetric_difference(set(g1.edges()))) == 1:
                        graph_space.add_edge(ind1, ind2)

        return graph_space, graph_space.subgraph(wflow), graph_space.subgraph(woflow)



import numpy as np
from typing import Union, List
import galois

__all__ = ["PauliOp"]

GF = galois.GF(2)


class PauliOp:
    """Class for representing Pauli operators as matrices and strings.

    Args
    ----
    op: Union[np.ndarray, str, List[str]]
        The Pauli operator to be represented. Can be a matrix, a string, or a list of strings.

    Examples
    --------
    Create a Pauli operator from a matrix

    .. ipython:: python

        op = mp.PauliOp(np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]]))
        print(op)

    Create a Pauli operator from a string

    .. ipython:: python

        op = mp.PauliOp('XIZ;ZII;IIZ;IZI')
        print(op)

    Create a Pauli operator from a list of strings

    .. ipython:: python

        op = mp.PauliOp(['XIZ', 'ZII', 'IIZ', 'IZI'])
        print(op)

    Group
    -----
    operators
    """

    def __init__(self, op: Union[np.ndarray, str, List[str]]):
        """Initialize a PauliOp object."""

        if isinstance(op, np.ndarray):
            if op.shape[1] % 2 != 0:
                raise ValueError(
                    "Tableau representation must have an even number of columns"
                )
            self.mat = GF(op)
            self.txt = self._mat_to_txt(op)

        elif isinstance(op, str):
            op = op.replace(" ", "")
            if op[-1] == ";":
                op = op[:-1]

            op_list = op.split(";")

            if not all([len(op) == len(op_list[0]) for op in op_list]):
                raise ValueError("All Pauli operators must be the same length")

            self.txt = op.replace(";", "\n")
            self.mat = self._txt_to_mat(op_list)

        elif isinstance(op, list):
            if not all([len(pauliop) == len(op[0]) for pauliop in op]):
                raise ValueError("All Pauli operators must be the same length")

            self.txt = "\n".join(op)
            self.mat = self._txt_to_mat(op)

        else:
            raise ValueError(
                "PauliOp must be initialized with a string or a numpy array"
            )

        if self.mat.shape[0] == 1:
            self._array = [self]
        else:
            self._array = [PauliOp(op) for op in self.txt.split("\n")]

    def __repr__(self):
        return self.txt

    def __getitem__(self, key):
        paulis = self._array[key]
        if isinstance(key, slice):
            return PauliOp(";".join([pauli.txt for pauli in paulis]))
        return paulis

    def _mat_to_txt(self, op):
        n_qubits = op.shape[1] // 2
        txt = ""
        for i in range(op.shape[0]):
            for j in range(op.shape[1] // 2):
                if op[i, j] == 1 and op[i, j + n_qubits] == 0:
                    txt += "X"
                elif op[i, j] == 0 and op[i, j + n_qubits] == 1:
                    txt += "Z"
                elif op[i, j] == 1 and op[i, j + n_qubits] == 1:
                    txt += "Y"
                else:
                    txt += "I"
                if j == n_qubits - 1 and i != op.shape[0] - 1:
                    txt += "\n"
        return txt

    def _txt_to_mat(self, op_list):
        n_qubits = len(op_list[0])
        n_ops = len(op_list)

        mat = np.zeros((n_ops, 2 * n_qubits), dtype=int)

        for i, op in enumerate(op_list):
            for j, char in enumerate(op):
                if char == "X":
                    mat[i, j] = 1
                elif char == "Z":
                    mat[i, j + n_qubits] = 1
                elif char == "Y":
                    mat[i, j] = 1
                    mat[i, j + n_qubits] = 1
        return GF(mat)

    def commutator(self, other):
        """Returns the commutator of two Pauli operators."""
        return self.mat + other.mat

    def symplectic_prod(self, other):
        """Returns the symplectic product of two Pauli operators."""
        x1 = self.mat[:, : self.mat.shape[1] // 2]
        x2 = other.mat[:, : other.mat.shape[1] // 2]
        z1 = self.mat[:, self.mat.shape[1] // 2 :]
        z2 = other.mat[:, other.mat.shape[1] // 2 :]
        return x1 @ z2.T + z1 @ x2.T

    def append(self, other):
        """Appends a Pauli operator to the end of another Pauli operator."""
        self.mat = np.vstack((self.mat, other.mat))
        self.txt = self._mat_to_txt(self.mat)



import numpy as np
from mentpy.gradients import estimate_gradient


def compute_gradient_variance(f, x, estimate_gradient, num_samples=10, **kwargs):
    """Compute the variance of the gradient of a function f."""
    grads = []
    for _ in range(num_samples):
        grads.append(estimate_gradient(f, x, **kwargs))
    grads = np.array(grads)
    return np.var(grads, axis=0)



"""This is the Flow module. It deals with the flow of a given graph state"""
import math
import numpy as np
import networkx as nx

from mentpy.states import GraphState
from typing import List
import warnings

import galois


def find_flow(graph: GraphState, input_nodes, output_nodes, sanity_check=True):
    r"""Finds the generalized flow of graph state if allowed.

    Implementation of https://arxiv.org/pdf/quant-ph/0603072.pdf.

    Returns
    -------
    The flow function ``flow`` and the partial order function.

    Examples
    --------
    Find the flow of a graph state :math:`|G\rangle`.

    .. ipython:: python

        g = nx.Graph()
        g.add_edges_from([(0,1), (1,2), (2,3), (3, 4)])
        state = mtp.MBQCGraph(g, input_nodes = [0], output_nodes = [4])
        flow, partial_order = mtp.find_flow(state)
        print("Flow of node 1: ", flow(1))
        print("Partial order: ", partial_order)

    Group
    -----
    states
    """
    n_input, n_output = len(input_nodes), len(output_nodes)
    inp = input_nodes
    outp = output_nodes
    if n_input != n_output:
        raise ValueError(
            f"Cannot find flow or gflow. Input ({n_input}) and output ({n_output}) nodes have different size."
        )

    update_labels = False
    # check if labels of graph are integers going from 0 to n-1 and if not, create a mapping
    if not all([i in graph.nodes for i in range(len(graph))]):
        mapping = {v: i for i, v in enumerate(graph.nodes)}
        inverse_mapping = {i: v for i, v in enumerate(graph.nodes)}
        # create a copy of the object state
        old_state = graph
        new_graph = nx.relabel_nodes(graph.copy(), mapping)
        inp, outp = [mapping[v] for v in input_nodes], [
            mapping[v] for v in output_nodes
        ]

        graph = GraphState(new_graph)
        update_labels = True

    tau = _build_path_cover(graph, inp, outp)
    if tau:
        f, P, L = _get_chain_decomposition(graph, inp, outp, tau)
        sigma = _compute_suprema(graph, inp, outp, f, P, L)

        if sigma is not None:
            int_flow = _flow_from_array(graph, inp, outp, f)
            vertex2index = {v: index for index, v in enumerate(inp)}

            def int_partial_order(x, y):
                return sigma[vertex2index[int(P[y])], int(x)] <= L[y]

            # if labels were updated, update them back
            if update_labels:
                graph = old_state
                flow = lambda v: inverse_mapping[int_flow(mapping[v])]
                partial_order = lambda x, y: int_partial_order(mapping[x], mapping[y])
            else:
                flow = int_flow
                partial_order = int_partial_order

            state_flow = (flow, partial_order)
            if sanity_check:
                if not check_if_flow(graph, inp, outp, flow, partial_order):
                    raise RuntimeError(
                        "Sanity check found that flow does not satisfy flow conditions."
                    )
            return state_flow

        else:
            warnings.warn(
                "The given state does not have a flow.", UserWarning, stacklevel=2
            )
            return None, None
    else:
        warnings.warn(
            "Could not find a flow for the given state.", UserWarning, stacklevel=2
        )
        return None, None


def _flow_from_array(graph: GraphState, input_nodes, output_nodes, f: List):
    """Create a flow function from a given array f"""

    def flow(v):
        if v in [v for v in graph.nodes() if v not in output_nodes]:
            return int(f[v])
        else:
            raise UserWarning(f"The node {v} is not in domain of the flow.")

    return flow


def _get_chain_decomposition(
    graph: GraphState, input_nodes, output_nodes, C: nx.DiGraph
):
    """Gets the chain decomposition"""
    P = np.zeros(len(graph))
    L = np.zeros(len(graph))
    f = {v: 0 for v in set(graph) - set(output_nodes)}
    for i in input_nodes:
        v, l = i, 0
        while v not in output_nodes:
            f[v] = int(next(C.successors(v)))
            P[v] = i
            L[v] = l
            v = int(f[v])
            l += 1
        P[v], L[v] = i, l
    return (f, P, L)


def _compute_suprema(graph: GraphState, input_nodes, output_nodes, f, P, L):
    """Compute suprema

    status: 0 if none, 1 if pending, 2 if fixed.
    """
    (sup, status) = _init_status(graph, input_nodes, output_nodes, P, L)
    for v in set(graph.nodes()) - set(output_nodes):
        if status[v] == 0:
            (sup, status) = _traverse_infl_walk(
                graph, input_nodes, output_nodes, f, sup, status, v
            )

        if status[v] == 1:
            return None

    return sup


def _traverse_infl_walk(
    graph: GraphState, input_nodes, output_nodes, f, sup, status, v
):
    """Compute the suprema by traversing influencing walks

    status: 0 if none, 1 if pending, 2 if fixed.
    """
    status[v] = 1
    vertex2index = {v: index for index, v in enumerate(input_nodes)}

    for w in list(graph.neighbors(f[v])) + [f[v]]:
        if w != v:
            if status[w] == 0:
                (sup, status) = _traverse_infl_walk(
                    graph, input_nodes, output_nodes, f, sup, status, w
                )
            if status[w] == 1:
                return (sup, status)
            else:
                for i in input_nodes:
                    if sup[vertex2index[i], v] > sup[vertex2index[i], w]:
                        sup[vertex2index[i], v] = sup[vertex2index[i], w]
    status[v] = 2
    return sup, status


def _init_status(graph: GraphState, input_nodes: List, output_nodes: List, P, L):
    """Initialize the supremum function

    status: 0 if none, 1 if pending, 2 if fixed.
    """
    sup = np.zeros((len(input_nodes), len(graph.nodes())))
    vertex2index = {v: index for index, v in enumerate(input_nodes)}
    status = np.zeros(len(graph.nodes()))
    for v in graph.nodes():
        for i in input_nodes:
            if i == P[v]:
                sup[vertex2index[i], v] = L[v]
            else:
                sup[vertex2index[i], v] = len(graph.nodes())

        status[v] = 2 if v in output_nodes else 0

    return sup, status


def _build_path_cover(graph: GraphState, input_nodes: List, output_nodes: List):
    """Builds a path cover

    status: 0 if 'fail', 1 if 'success'
    """
    fam = nx.DiGraph()
    visited = np.zeros(graph.number_of_nodes())
    iter = 0
    for i in input_nodes:
        iter += 1
        (fam, visited, status) = _augmented_search(
            graph, input_nodes, output_nodes, fam, iter, visited, i
        )
        if not status:
            return status

    if not len(set(graph.nodes) - set(fam.nodes())):
        return fam

    return 0


def _augmented_search(
    graph: GraphState,
    input_nodes: List,
    output_nodes: List,
    fam: nx.DiGraph,
    iter: int,
    visited,
    v,
):
    """Does an augmented search

    status: 0 if 'fail', 1 if 'success'
    """
    visited[v] = iter
    if v in output_nodes:
        return (fam, visited, 1)
    if (
        (v in fam.nodes())
        and (v not in input_nodes)
        and (visited[next(fam.predecessors(v))] < iter)
    ):
        (fam, visited, status) = _augmented_search(
            graph,
            input_nodes,
            output_nodes,
            fam,
            iter,
            visited,
            next(fam.predecessors(v)),
        )
        if status:
            fam = fam.remove_edge(next(fam.predecessors(v)), v)
            return (fam, visited, 1)

    for w in graph.neighbors(v):
        if (visited[w] < iter) and (w not in input_nodes) and (not fam.has_edge(v, w)):
            if w not in fam.nodes():
                (fam, visited, status) = _augmented_search(
                    graph, input_nodes, output_nodes, fam, iter, visited, w
                )
                if status:
                    fam.add_edge(v, w)
                    return (fam, visited, 1)
            elif visited[next(fam.predecessors(w))] < iter:
                (fam, visited, status) = _augmented_search(
                    graph, fam, iter, visited, next(fam.predecessors(w))
                )
                if status:
                    fam.remove_edge(next(fam.predecessors(w)), w)
                    fam.add_edge(v, w)
                    return (fam, visited, 1)

    return (fam, visited, 0)


def check_if_flow(
    graph: GraphState, input_nodes: List, output_nodes: List, flow, partial_order
) -> bool:
    """Checks if flow satisfies conditions on state."""
    conds = True
    for i in [v for v in graph.nodes() if v not in output_nodes]:
        nfi = list(graph.neighbors(flow(i)))
        c1 = i in nfi
        c2 = partial_order(i, flow(i))
        c3 = math.prod([partial_order(i, k) for k in set(nfi) - {i}])
        conds = conds * c1 * c2 * c3
        if not c1:
            print(f"Condition 1 failed for node {i}. {i} not in {nfi}")
        if not c2:
            print(f"Condition 2 failed for node {i}. {i} ≮ {flow(i)}")
        if not c3:
            print(f"Condition 3 failed for node {i}.")
            for k in set(nfi) - {i}:
                if not partial_order(i, k):
                    print(f"{i} ≮ {k}")
    return conds


### This section implements causal flow


def causal_flow(graph: GraphState, input_nodes, output_nodes) -> object:
    """Finds the causal flow of a ``MBQCGraph`` if it exists.
    Retrieved from https://arxiv.org/pdf/0709.2670v1.pdf.
    """
    l = {}

    for v in output_nodes:
        l[v] = 0

    result, flow = causal_flow_aux(
        graph,
        set(input_nodes),
        set(output_nodes),
        set(output_nodes) - set(input_nodes),
        1,
        l,
    )

    if not result:
        raise UserWarning("No causal flow exists for this graph.")

    # return flow and partial order
    return lambda x: flow[x], lambda u, v: l[u] > l[v]


def causal_flow_aux(graph: GraphState, inputs, outputs, C, k, l) -> object:
    """Aux function for causal_flow"""
    V = set(graph.nodes())
    out_prime = set()
    C_prime = set()
    g = {}

    for _, v in enumerate(C):
        # get intersection of neighbors of v and (V \ output nodes
        intersection = set(graph.neighbors(v)) & (V - outputs)
        if len(intersection) == 1:
            u = intersection.pop()
            g[u] = v
            l[v] = k
            out_prime.add(u)
            C_prime.add(v)

    if len(out_prime) == 0:
        if set(outputs) == V:
            return True, l
        else:
            return False, l
    else:
        return causal_flow_aux(
            graph,
            inputs,
            outputs | out_prime,
            (C - C_prime) | ((out_prime & V) - inputs),
            k + 1,
            l,
        )


### This section is for GFlow ###

# def gflow(state: MBQCGraph) -> object:
#     """Finds the generalized flow of a ``MBQCGraph`` if it exists.
#    Retrieved from https://arxiv.org/pdf/0709.2670v1.pdf.
#    """
#     gamma = nx.adjacency_matrix(state.graph).toarray()
#     l = {}
#     g = {}

#     for v in state.output_nodes:
#         l[v] = 0


#     result, flow, g = gflowaux(set(state.graph.nodes()), gamma, set(state.input_nodes), set(state.output_nodes), 1, l, g)

#     return lambda x: flow[x], lambda u, v: l[u] > l[v], lambda u: g[u]

# def gflowaux(V, gamma, inputs, outputs, k, l, g) -> object:
#     """Aux function for gflow"""
#     GF = galois.GF(2)
#     node_mapping = {i: node for i, node in enumerate(V)}

#     C = set()
#     for u in V-outputs:
#         # submatrix = gamma[np.ix_(list(V-outputs), list(outputs - inputs))]
#         # get submatrix with rows corresponding to V-outputs and columns corresponding to outputs - inputs
#         # but use the node_mapping to get the correct indices
#         submatrix = np.zeros((len(V-outputs), len(outputs - inputs)), dtype=int)
#         for i, v in enumerate(V-outputs):
#             for j, w in enumerate(outputs - inputs):
#                 submatrix[i, j] = gamma[node_mapping[v], node_mapping[w]]

#         iu = np.zeros(len(V-outputs), dtype=int)
#         iu[list(V-outputs).index(u)] = 1
#         # use node_mapping to get the correct index

#         submatrix, iu = GF(submatrix), GF(np.array([iu]).T)
#         #solve submatrix @ X = iu using galois in mod2
#         sys = np.hstack((submatrix, iu))
#         row_reduced = sys.row_reduce()

#         # FILL HERE TO GET SOLUTION


#         # check if solution is a valid solution
#         if np.linalg.norm(submatrix @ solution - iu) <= 1e-5:
#             l[u] = k
#             C.add(u)
#             g[u] = solution

#     if len(C) == 0:
#         return True, l, g
#     else:
#         return gflowaux(V, gamma, inputs, outputs | C, k+1, l, g)


# def _check_if_gflow(state: MBQCGraph, gflow, partial_order) -> bool:
#     """Checks if gflow satisfies conditions on state."""


def find_gflow(graph: GraphState, input_nodes, output_nodes) -> object:
    """Finds the generalized flow of a ``MBQCGraph`` if it exists.
    Retrieved from https://arxiv.org/pdf/0709.2670v1.pdf.
    """
    gamma = nx.adjacency_matrix(graph).toarray()
    l = {}
    g = {}

    for v in output_nodes:
        l[v] = 0

    result, flow, g = gflowaux(
        graph, gamma, set(input_nodes), set(output_nodes) - set(input_nodes), 1, l, g
    )

    if result == False:
        warnings.warn("No gflow exists for this graph.", UserWarning, stacklevel=2)

    return lambda x: flow[x], lambda u, v: l[u] > l[v], lambda u: g[u]


def gflowaux(graph: GraphState, gamma, inputs, outputs, k, l, g) -> object:
    """Aux function for gflow"""
    out_prime = set()
    GF = galois.GF(2)
    V = set(graph.nodes())
    C = set()
    for u in V - outputs:
        submatrix = GF(np.zeros((len(V - outputs), len(outputs)), dtype=int))
        b = GF(np.zeros((len(V - outputs), 1), dtype=int))
        b[list(V - outputs).index(u)] = 1
        syst = np.hstack((submatrix, b))
        row_reduced = GF(syst).row_reduce()
        solution = GF(np.zeros((len(outputs), 1), dtype=int))
        for i, v in enumerate(outputs):
            if row_reduced[i, i] == 1:
                solution[i] = row_reduced[i, -1]

        if np.linalg.norm(submatrix @ solution - b) <= 1e-5:
            l[u] = k
            C.add(u)
            g[u] = solution

    if len(C) == 0:
        if set(outputs) == V:
            return True, l, g
        else:
            return False, l, g

    else:
        return gflowaux(graph, gamma, inputs, outputs | C, k + 1, l, g)



from setuptools import setup, find_packages

version = '0.1.0a2'

setup(
    name='mentpy',
    version=version,
    author='Luis Mantilla',
    author_email='luismantilla99@outlook.com',
    description='A Python library for simulating MBQC circuits',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/BestQuark/mentpy',
    packages=find_packages(),
)



import numpy as np
import cirq

from mentpy import MBQCState


def generate_random_input_states(mbqc_circuit: MBQCState, n_samples: int = 1) -> list:
    r"""Generates ``n_samples`` random states for the given ``mbqc_circuit``
    sampled using the Haar measure."""

    if n_samples < 1:
        raise UserWarning(
            f"n_samples expected to be greater than or equal to 1, "
            f"but {n_samples} was given."
        )

    n_qubits = len(mbqc_circuit.input_nodes)
    return [generate_haar_random_states(n_qubits) for _ in range(n_samples)]


def _generate_haar_random_state(n_qubits: int) -> np.ndarray:
    r"""Makes one Haar random state over n_qubits"""

    zero_list = n_qubits * [cirq.KET_ZERO.state_vector()]
    ket_zeros = cirq.kron(*zero_list)
    haar_random_u = cirq.testing.random_special_unitary(dim=int(2**n_qubits))
    return (haar_random_u @ ket_zeros.T).T[0]


def generate_haar_random_states(n_qubits: int, n_samples: int = 1) -> np.ndarray:
    r"""Makes one Haar random state over n_qubits"""

    if n_samples == 1:
        return _generate_haar_random_state(n_qubits)
    else:
        return [_generate_haar_random_state(n_qubits) for _ in range(n_samples)]


def random_special_unitary(n_qubits: int):
    """Returns a random special unitary in ``n_qubits`` sampled from the Haar distribution."""
    return cirq.testing.random_special_unitary(dim=int(2**n_qubits))


def random_train_test_data_for_unitary(
    unitary: np.ndarray, n_samples: int, test_size: float = 0.3
) -> tuple:
    r"Return random training and test data (input, target) for a given unitary gate ``unitary``."
    n_qubits = int(np.log2(unitary.shape[0]))
    random_inputs = generate_haar_random_states(n_qubits, n_samples=n_samples)
    random_targets = [(unitary @ st.T).T for st in random_inputs]
    return _train_test_split(random_inputs, random_targets, test_size=test_size)


def _train_test_split(inputs, targets, test_size: float = 0.3) -> tuple:
    r"Split the data into training and test sets."
    n_samples = len(inputs)
    n_test_samples = int(n_samples * test_size)
    n_train_samples = n_samples - n_test_samples
    return (inputs[:n_train_samples], targets[:n_train_samples]), (
        inputs[n_train_samples:],
        targets[n_train_samples:],
    )


def random_training_data_for_unitary(
    mbqc_circuit: MBQCState, n_samples: int, unitary: np.ndarray
) -> tuple:
    r"Return random training data (input, target) for a given unitary gate ``unitary``."
    random_inputs = generate_random_input_states(mbqc_circuit, n_samples=n_samples)
    random_targets = [(unitary @ st.T).T for st in random_inputs]
    return random_inputs, random_targets



import numpy as np
import cirq
import networkx as nx

from scipy.spatial import distance
from scipy.special import kl_div, rel_entr

import pennylane as qml

from mentpy.states import MBQCState
from mentpy.simulators import PatternSimulator
from mentpy.utils.lc_equivalence import are_lc_equivalent
from mentpy.utils.generate_data import generate_random_input_states
from mentpy.utils.flow_space import FlowSpace


def haar_probability_density_of_fidelities(F: float, n_qubits: int):
    r"""Returns the probability density function of fidelities
    :math:`P_{Haar}(F) = (N-1)(1-F)^{N-2}` where :math:`N = 2^{n}` is
    the dimension of the Hilbert space.

    Args
    ----
    F (float): Fidelity. Must be between :math:`0` and :math:`1`.
    n_qubits (int): Number of qubits. Must be greater than or equal to 1.
    n_bins (int): Number of bins used in the histogram

    Returns
    -------
    :math:`P_{Haar}(F) = (N-1)(1-F)^{N-2}`  where  :math:`N = 2^{n}`
    """
    N = int(2**n_qubits)
    return (N - 1) * ((1 - F) ** (N - 2))


def expressivity_with_histogram(
    graph_state_circuit: MBQCState,
    n_samples: int = 10000,
    n_bins: int = 1000,
    method: str = "KL",
):
    r"""Returns the expressivity calculated using the Kullback-Leiber entropy

    Args
    ----
    graph_state_circuit: Graph state circuit for which we will calculate the expressivity
    n_samples: Number of samples that we will use to estimate the expressivity
    n_bins: Number of bins of the histogram
    method: 'KL' for Kullback-Leibler divergence or 'RE' for relative entropy.

    Returns
    -------
    expressivity (float): Expressivity measure for graph_state_circuit
    """
    expressivity = -1
    # TODO: Add covering number method to calculate expressivity (?)
    if method not in ["KL", "RE", "JS"]:
        raise UserWarning("Unsupported method for calculating expressivity")

    samples_from_circuit = sample_probability_density_of_fidelities(
        graph_state_circuit, n_samples=n_samples
    )
    digitized_data, edges_hist = np.histogram(
        samples_from_circuit, bins=n_bins, density=True, range=(0, 1)
    )

    midpoints = 0.5 * (edges_hist[:-1] + edges_hist[1:])
    haar_prob_fun = lambda fid: haar_probability_density_of_fidelities(
        fid, len(graph_state_circuit.output_nodes)
    )

    digitized_haar = haar_prob_fun(np.array(midpoints))

    discrete_prob_data = digitized_data * np.diff(edges_hist)
    discrete_prob_haar = digitized_haar * np.diff(edges_hist)

    if method == "KL":
        expressivity = sum(kl_div(discrete_prob_data, discrete_prob_haar))
    elif method == "RE":
        expressivity = sum(rel_entr(discrete_prob_data, discrete_prob_haar))
    elif method == "JS":
        expressivity = distance.jensenshannon(discrete_prob_data, discrete_prob_haar)

    return expressivity


def sample_probability_density_of_fidelities(
    graph_state_circuit: MBQCState, n_samples=1000, backend="pennylane"
):
    r"""Calculates samples of the probability of fidelities of the given graph state circuit"""

    pattern_simulator = PatternSimulator(graph_state_circuit)
    random_input_states = generate_random_input_states(
        pattern_simulator.state, n_samples
    )
    fidelities = []

    if backend == "pennylane":
        qmlcircuit = pattern_simulator.graphstate_to_circuit()
        for random_st in random_input_states:
            random_pattern = (
                2 * np.pi * np.random.rand(pattern_simulator.max_measure_number)
            )
            final_state = qmlcircuit(random_pattern, output="density")
            fid = qml.math.fidelity(np.kron(random_st.T, random_st), final_state)
            fidelities.append(fid)

    elif backend == "cirq":
        for random_st in random_input_states:
            random_pattern = (
                2 * np.pi * np.random.rand(pattern_simulator.max_measure_number)
            )
            pattern_simulator.reset(input_state=random_st)

            _ = pattern_simulator.measure_pattern(random_pattern)
            final_state = pattern_simulator.current_sim_state
            fidelities.append(cirq.fidelity(random_st, final_state))

    else:
        print(f"Unsupported backend {backend}")

    return fidelities


def digraph_expressivity_of_flow_space(flow_space: FlowSpace, method="KL", **kwargs):
    """Returns digraph given the expressivity of a ``FlowSpace`` object.

    Args
    ----
        flow_space (FlowSpace): Graph for which we will calculate the expressivity
        method (str): 'KL' for Kullback-Leiber entropy, 'RE' for relative entropy.
        n_samples (int): Number of samples to calculate fidelity histogram. 10000 is default.
        n_bins (int): Number of bins of fidelity histogram. 1000 is default.
    """

    expr_digraph = nx.DiGraph()
    for node in flow_space.flow_graph_space.nodes():
        graph_circuit = flow_space.flow_graph_space.nodes[node]["mbqc_circuit"]
        expr = expressivity_with_histogram(graph_circuit, method=method, **kwargs)
        expr_digraph.add_node(node, expr=expr, mbqc_circuit=graph_circuit)

    for edge in flow_space.flow_graph_space.edges():
        i, j = edge
        expri, exprj = expr_digraph.nodes[i]["expr"], expr_digraph.nodes[j]["expr"]
        if expri <= exprj:
            expr_digraph.add_edge(j, i)
        else:
            expr_digraph.add_edge(i, j)

    return expr_digraph


def lc_cluster_flowspace(deg_graph, sanity_check=True):
    """Cluster flow graph in lc_equivalent graphs."""
    independent_groups = {}
    independent_groups_ind = {}
    for node in deg_graph.nodes():
        curr_g = deg_graph.nodes[node]["mbqc_circuit"].graph
        if not independent_groups:
            independent_groups[0] = [curr_g]
            independent_groups_ind[0] = [node]
        else:
            added = False
            for k, g in independent_groups.items():

                if are_lc_equivalent(g[0], curr_g)[0]:
                    added = True
                    independent_groups[k].append(curr_g)
                    independent_groups_ind[k].append(node)
                    if sanity_check:
                        for n in g:
                            if not are_lc_equivalent(n, curr_g)[0]:
                                raise RuntimeError(
                                    "Graph not lc equivalent with all group"
                                )
            if not added:
                new_key = max(independent_groups) + 1
                independent_groups[new_key] = [curr_g]
                independent_groups_ind[new_key] = [node]

    return independent_groups_ind, independent_groups


def draw_digraph_flow_space(digraph_expr: nx.digraph, **kwargs):
    """Draws the expressivity digraph of flow space"""
    pospos = nx.spring_layout(digraph_expr.to_undirected())
    nx.draw(digraph_expr, node_size=55, pos=pospos, **kwargs)
